##
# Copyright (c) Microsoft Corporation.
# Licensed under the MIT License.
##
try:
    import cirq
except ImportError:
    raise ImportError(
    "Missing optional 'cirq' dependencies. \
To install run: pip install azure-quantum[cirq]"
)

from azure.quantum import Workspace
from azure.quantum.job.base_job import DEFAULT_TIMEOUT
from azure.quantum.cirq.targets import * 

from typing import Optional, Union, List, TYPE_CHECKING

if TYPE_CHECKING:
    from azure.quantum.cirq.targets import Target as CirqTarget
    from azure.quantum.cirq.job import Job as CirqJob
    from cirq_ionq import Job as CirqIonqJob

DEFAULT_JOB_NAME = "cirq-job"
CIRQ_USER_AGENT = "azure-quantum-cirq"


class AzureQuantumService:
    """
    Class for interfacing with the Azure Quantum service
    using Cirq quantum circuits
    """
    def __init__(
        self,
        workspace: Workspace = None,
        default_target: Optional[str] = None,
        **kwargs
    ):
        """AzureQuantumService class

        :param workspace: Azure Quantum workspace. If missing it will create a new Workspace passing `kwargs` to the constructor. Defaults to None. 
        :type workspace: Workspace, optional
        :param default_target: Default target name, defaults to None
        :type default_target: Optional[str], optional
        """
        if workspace is None:
            workspace = Workspace(**kwargs)

        workspace.append_user_agent(CIRQ_USER_AGENT)

        self._workspace = workspace
        self._default_target = default_target

    @property
    def _target_factory(self):
        from azure.quantum.target.target_factory import TargetFactory
        from azure.quantum.cirq.targets import Target, DEFAULT_TARGETS

        target_factory = TargetFactory(
            base_cls=Target,
            workspace=self._workspace,
            default_targets=DEFAULT_TARGETS
        )

        return target_factory

    def targets(
        self,
        name: str = None,
        provider_id: str = None,
        **kwargs
    ) -> Union["CirqTarget", List["CirqTarget"]]:
        """Get all quantum computing targets available in the Azure Quantum Workspace.

        :param name: Target name, defaults to None
        :type name: str, optional
        :return: Target instance or list thereof
        :rtype: Union[Target, List[Target]]
        """
        return self._target_factory.get_targets(
            name=name,
            provider_id=provider_id
        )

    def get_target(self, name: str = None, **kwargs) -> "CirqTarget":
        """Get target with the specified name

        :param name: Target name
        :type name: str
        :return: Cirq target
        :rtype: CirqTarget
        """
        if name is None:
            if self._default_target is None:
                raise ValueError("No default target specified for job.")
            return self.targets(name=self._default_target, **kwargs)

        if isinstance(name, str):
            return self.targets(name=name, **kwargs)

    def get_job(self, job_id: str, *args, **kwargs) -> Union["CirqJob", "CirqIonqJob"]:
        """Get Cirq Job by job ID

        :param job_id: Job ID
        :type job_id: str
        :return: Job
        :rtype: azure.quantum.cirq.Job or cirq_ionq.Job
        """
        job = self._workspace.get_job(job_id=job_id)
        target : CirqTarget = self._target_factory.create_target(
            provider_id=job.details.provider_id,
            name=job.details.target
        )
        return target._to_cirq_job(azure_job=job, *args, **kwargs)

    def create_job(
        self,
        program: cirq.Circuit,
        repetitions: int,
        name: str = DEFAULT_JOB_NAME,
        target: str = None,
        param_resolver: cirq.ParamResolverOrSimilarType = cirq.ParamResolver({})
    ) -> Union["CirqJob", "CirqIonqJob"]:
        """Create job to run the given `cirq` program in Azure Quantum

        :param program: Cirq program or circuit
        :type program: cirq.Circuit
        :param repetitions: Number of measurements 
        :type repetitions: int
        :param name: Program name
        :type name: str
        :param target: Target name
        :type target: str
        :param param_resolver: Parameter resolver for cirq program
        :type param_resolver: cirq.ParamResolverOrSimilarType
        :return: Job
        :rtype: azure.quantum.cirq.Job
        """
        # Get target
        _target = self.get_target(name=target)
        if not _target:
            target_name = target or self._default_target
            raise RuntimeError(f"Could not find target '{target_name}'. \
Please make sure the target name is valid and that the associated provider is added to your Workspace. \
To add a provider to your quantum workspace on the Azure Portal, \
see https://aka.ms/AQ/Docs/AddProvider")
        # Resolve parameters
        resolved_circuit = cirq.resolve_parameters(program, param_resolver)
        # Submit job to Azure
        return _target.submit(
            program=resolved_circuit,
            repetitions=repetitions,
            name=name
        )

    def estimate_cost(
        self,
        program: cirq.Circuit,
        repetitions: int,
        target: str = None,
        param_resolver: cirq.ParamResolverOrSimilarType = cirq.ParamResolver({}),
        **kwargs
    ):
        # Resolve parameters
        resolved_circuit = cirq.resolve_parameters(program, param_resolver)
        target = self.get_target(name=target)
        return target.estimate_cost(
            program=resolved_circuit,
            repetitions=repetitions,
            **kwargs
        )

    def run(
        self,
        program: cirq.Circuit,
        repetitions: int,
        target: str = None,
        name: str = DEFAULT_JOB_NAME,
        param_resolver: cirq.ParamResolverOrSimilarType = cirq.ParamResolver({}),
        seed: cirq.RANDOM_STATE_OR_SEED_LIKE = None,
        timeout_seconds: int = DEFAULT_TIMEOUT,
    ) -> cirq.Result:
        """Run Cirq circuit on specified target, if target not specified then it runs on the default target

        :param program: Cirq program or circuit
        :type program: cirq.Circuit
        :param repetitions: Number of measurement repetitions
        :type repetitions: int
        :param target: Target name, defaults to default_target
        :type target: str, optional
        :param name: Program name, defaults to "cirq-job"
        :type name: str, optional
        :param param_resolver: Cirq parameters, defaults to cirq.ParamResolver({})
        :type param_resolver: cirq.ParamResolverOrSimilarType, optional
        :param seed: Random seed for simulator results, defaults to None
        :type seed: cirq.RANDOM_STATE_OR_SEED_LIKE, optional
        :param timeout_seconds: Timeout in seconds, defaults to None
        :type timeout_seconds: int, optional
        :return: Measurement results
        :rtype: cirq.Result
        """
        job = self.create_job(
            program=program,
            repetitions=repetitions,
            name=name,
            target=target,
            param_resolver=param_resolver
        )
        # Get raw job results
        try:
            result = job.results(timeout_seconds=timeout_seconds)
        except RuntimeError as e:
            # Catch errors from cirq_ionq.Job.results
            if "Job was not completed successful. Instead had status: " in str(e):
                raise TimeoutError(f"The wait time has exceeded {timeout_seconds} seconds. \
Job status: '{job.status()}'.")
            else:
                raise e

        # Convert to Cirq Result
        target = self.get_target(name=target)
        return target._to_cirq_result(
            result=result,
            param_resolver=param_resolver,
            seed=seed
        )



# coding=utf-8
##
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License.
##

from .term import *
from .problem import *
from .streaming_problem import *
from .online_problem import *
from azure.quantum.target.solvers import *
from azure.quantum.target.microsoft.qio import (
    ParallelTempering,
    PopulationAnnealing,
    QuantumMonteCarlo,
    SimulatedAnnealing,
    SubstochasticMonteCarlo,
    Tabu,
)



##
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License.
##
import json
from json.decoder import JSONDecodeError
import logging
import os
import time
from aiofile import async_open

from azure.identity import CredentialUnavailableError
from azure.core.credentials import AccessToken

try:
    from typing import TYPE_CHECKING
except ImportError:
    TYPE_CHECKING = False

if TYPE_CHECKING:
    # pylint:disable=unused-import,ungrouped-imports
    from typing import Any

_LOGGER = logging.getLogger(__name__)

_TOKEN_FILE_ENV_VARIABLE = "AZURE_QUANTUM_TOKEN_FILE"


class _TokenFileCredential(object):
    """
    Implements a custom TokenCredential to use a local file as the source for an AzureQuantum token.

    It will only use the local file if the AZURE_QUANTUM_TOKEN_FILE environment variable is set, and references
    an existing json file that contains the access_token and expires_on timestamp in milliseconds.

    If the environment variable is not set, the file does not exist, or the token is invalid in any way (expired, for example),
    then the credential will throw CredentialUnavailableError, so that _ChainedTokenCredential can fallback to other methods.
    """
    def __init__(self, **kwargs):
        # type: (**Any) -> None
        self.token_file = os.environ.get(_TOKEN_FILE_ENV_VARIABLE)
        if self.token_file:
            _LOGGER.debug("Using provided token file location: {}".format(self.token_file))
        else:
            _LOGGER.debug("No token file location provided for {} environment variable.".format(_TOKEN_FILE_ENV_VARIABLE))

    async def get_token(self, *scopes, **kwargs):  # pylint:disable=unused-argument
        # type: (*str, **Any) -> AccessToken
        """Request an access token for `scopes`.
        This method is called automatically by Azure SDK clients.
        :param str scopes: desired scopes for the access token. This method only returns tokens for the https://quantum.microsoft.com/.default scope.
        :raises ~azure.identity.CredentialUnavailableError: when failing to get token. The exception has a
          `message` attribute with the error message.
        """
        if not self.token_file:
            raise CredentialUnavailableError(message="Token file location not set.")

        if not os.path.isfile(self.token_file):
            raise CredentialUnavailableError(message="Token file at {} does not exist.".format(self.token_file))

        try:
            token = await self._parse_token_file(self.token_file)
        except JSONDecodeError:
            raise CredentialUnavailableError(message="Failed to parse token file: Invalid JSON.")
        except KeyError as e:
            raise CredentialUnavailableError(message="Failed to parse token file: Missing expected value: " + str(e))
        except Exception as e:
            raise CredentialUnavailableError(message="Failed to parse token file: " + str(e))

        if token.expires_on <= time.time():
            raise CredentialUnavailableError(message="Token already expired at {}".format(time.asctime(time.gmtime(token.expires_on))))

        return token

    async def _parse_token_file(self, path):
        # type: (*str) -> AccessToken
        async with async_open(path, "r") as file:
            data = json.load(file)
            expires_on = int(data["expires_on"]) / 1000  # Convert ms to seconds, since python time.time only handles epoch time in seconds
            token = AccessToken(data["access_token"],  expires_on)
            return token



##
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License.
##
import io
import json

from typing import Any, Dict, List

from azure.quantum.target.target import Target
from azure.quantum.job.job import Job
from azure.quantum.workspace import Workspace
from azure.quantum._client.models import CostEstimate, UsageEvent

def int_to_bitstring(k: int, num_qubits: int, measured_qubit_ids: List[int]):
    # flip bitstring to convert to little Endian
    bitstring = format(int(k), f"0{num_qubits}b")[::-1]
    # flip bitstring to convert back to big Endian
    return "".join([bitstring[n] for n in measured_qubit_ids])[::-1]


class IonQ(Target):
    """IonQ target."""
    target_names = (
        "ionq.qpu",
        "ionq.simulator",
    )

    def __init__(
        self,
        workspace: Workspace,
        name: str = "ionq.simulator",
        input_data_format: str = "ionq.circuit.v1",
        output_data_format: str = "ionq.quantum-results.v1",
        provider_id: str = "IonQ",
        content_type: str = "application/json",
        encoding: str = "",
        **kwargs
    ):
        super().__init__(
            workspace=workspace,
            name=name,
            input_data_format=input_data_format,
            output_data_format=output_data_format,
            provider_id=provider_id,
            content_type=content_type,
            encoding=encoding,
            **kwargs
        )

    @staticmethod
    def _encode_input_data(data: Dict[Any, Any]) -> bytes:
        stream = io.BytesIO()
        data = json.dumps(data)
        stream.write(data.encode())
        return stream.getvalue()

    def submit(
        self,
        circuit: Dict[str, Any],
        name: str = "ionq-job",
        num_shots: int = None,
        input_params: Dict[str, Any] = None,
        **kwargs
    ) -> Job:
        """Submit an IonQ circuit (JSON format)

        :param circuit: Quantum circuit in IonQ JSON format (for examples,
            see: https://docs.ionq.com/#section/Sample-JSON-Circuits)
        :type circuit: Dict[str, Any]
        :param name: Job name
        :type name: str
        :param num_shots: Number of shots, defaults to None
        :type num_shots: int
        :param input_params: Optional input params dict
        :type input_params: Dict[str, Any]
        :return: Azure Quantum job
        :rtype: Job
        """
        if input_params is None:
            input_params = {}
        if num_shots is not None:
            input_params = input_params.copy()
            input_params["shots"] = num_shots

        return super().submit(
            input_data=circuit,
            name=name,
            input_params=input_params,
            **kwargs
        )

    def estimate_cost(
        self,
        circuit: Dict[str, Any],
        num_shots: int,
        price_1q: float=0.00003,
        price_2q: float=0.0003,
        min_price: float=1.0
    ) -> CostEstimate:
        """Estimate the cost of submittng a circuit to IonQ targets.
        Optionally, you can provide the number of gate and measurement operations
        manually.
        The actual price charged by the provider may differ from this calculation.
        
        Specify pricing details for your area to get most accurate results.
        By default, this function charges price_1q=0.00003 USD for a single-qubit gate,
        price_2q=0.0003 USD for a two-qubit gate with a total minimum price of $1.-
        per circuit.

        For the most current pricing details, see
        https://docs.microsoft.com/azure/quantum/provider-ionq#pricing
        Or find your workspace and view pricing options in the "Provider" tab
        of your workspace: http://aka.ms/aq/myworkspaces

        :param circuit: Quantum circuit in IonQ JSON format (for examples,
            see: https://docs.ionq.com/#section/Sample-JSON-Circuits)
        :type circuit: Dict[str, Any]
        :param num_shots: Number of shots, defaults to None
        :type num_shots: int
        :param price_1q: The price of running a single-qubit gate
            for one shot, defaults to 0.00003
        :type price_1q: float, optional
        :param price_2q: The price of running a double-qubit gate
            for one shot, defaults to 0.0003
        :type price_2q: float, optional
        :param min_price: The minimum price for running a job, defaults to 1.0
        :type min_price: float, optional
        """
        def is_1q_gate(gate: Dict[str, Any]):
            return "controls" not in gate and "control" not in gate

        def is_multi_q_gate(gate):
            return "controls" in gate or "control" in gate

        def num_2q_gates(gate):
            controls = gate.get("controls")
            if controls is None or len(controls) == 1:
                # Only one control qubit
                return 1
            # Multiple control qubits
            return 6 * (len(controls) - 2)

        gates = circuit.get("circuit", [])
        N_1q = sum(map(is_1q_gate, gates))
        N_2q = sum(map(num_2q_gates, filter(is_multi_q_gate, gates)))

        if self.name == "ionq.simulator": 
            price = 0.0
        else:
            price = (price_1q * N_1q + price_2q * N_2q) * num_shots
            price = max(price, min_price)

        return CostEstimate(
            events = [
                UsageEvent(
                    dimension_id="gs1q",
                    dimension_name="1Q Gate Shot",
                    measure_unit="1q gate shot",
                    amount_billed=0.0,
                    amount_consumed=N_1q * num_shots,
                    unit_price=0.0
                ),
                UsageEvent(
                    dimension_id="gs2q",
                    dimension_name="2Q Gate Shot",
                    measure_unit="2q gate shot",
                    amount_billed=0.0,
                    amount_consumed=N_2q * num_shots,
                    unit_price=0.0
                )
            ],
            currency_code="USD",
            estimated_total=price
        )



##
# Copyright (c) Microsoft Corporation.
# Licensed under the MIT License.
##
from typing import TYPE_CHECKING
from azure.quantum import __version__
from azure.quantum.qiskit.job import AzureQuantumJob
from azure.quantum.target.ionq import IonQ

try:
    from qiskit.providers import BackendV1 as Backend
    from qiskit.providers.models import BackendConfiguration
    from qiskit.providers import Options
    from qiskit.qobj import Qobj, QasmQobj

    from qiskit_ionq.helpers import ionq_basis_gates, qiskit_circ_to_ionq_circ
except ImportError:
    raise ImportError(
    "Missing optional 'qiskit' dependencies. \
To install run: pip install azure-quantum[qiskit]"
)

if TYPE_CHECKING:
    from azure.quantum.qiskit import AzureQuantumProvider

import logging
logger = logging.getLogger(__name__)

__all__ = ["IonQBackend", "IonQQPUBackend", "IonQSimulatorBackend"]


class IonQBackend(Backend):
    """Base class for interfacing with an IonQ backend in Azure Quantum"""
    backend_name = None

    @classmethod
    def _default_options(cls):
        return Options(shots=500)

    def _job_metadata(self, circuit, meas_map):
        return {
            "qiskit": True,
            "name": circuit.name,
            "num_qubits": circuit.num_qubits,
            "meas_map": meas_map,
        }

    @staticmethod
    def _translate_circuit(circuit, **kwargs):
        ionq_circ, _, meas_map = qiskit_circ_to_ionq_circ(circuit)
        input_data = {
            "qubits": circuit.num_qubits,
            "circuit": ionq_circ,
        }

        return input_data, meas_map

    def estimate_cost(self, circuit, shots):
        """Estimate the cost for the given circuit."""
        input_data, _ = self._translate_circuit(circuit)
        workspace = self.provider().get_workspace()
        target = workspace.get_targets(self.name())
        return target.estimate_cost(input_data, num_shots=shots)

    def run(self, circuit, **kwargs):
        """Submits the given circuit to run on an IonQ target."""        
        # Some Qiskit features require passing lists of circuits, so unpack those here.
        # We currently only support single-experiment jobs.
        if isinstance(circuit, (list, tuple)):
            if len(circuit) > 1:
                raise NotImplementedError("Multi-experiment jobs are not supported!")
            circuit = circuit[0]

        # If the circuit was created using qiskit.assemble,
        # disassemble into QASM here
        if isinstance(circuit, QasmQobj) or isinstance(circuit, Qobj):
            from qiskit.assembler import disassemble
            circuits, run, _ = disassemble(circuit)
            circuit = circuits[0]
            if kwargs.get("shots") is None:
                # Note that the default number of shots for QObj is 1024
                # unless the user specifies the backend.
                kwargs["shots"] = run["shots"]

        input_data, meas_map = self._translate_circuit(circuit, **kwargs)
        input_data = IonQ._encode_input_data(input_data)

        # Options are mapped to input_params
        # Take also into consideration options passed in the kwargs, as the take precedence
        # over default values:
        input_params = vars(self.options)
        for opt in kwargs.copy():
            if opt in input_params:
                input_params[opt] = kwargs.pop(opt)

        logger.info(f"Submitting new job for basckend {self.name()}")
        job = AzureQuantumJob(
            backend=self,
            name=circuit.name,
            target=self.name(),
            input_data=input_data,
            blob_name="inputData",
            content_type="application/json",
            provider_id="ionq",
            input_data_format="ionq.circuit.v1",
            output_data_format="ionq.quantum-results.v1",
            input_params = input_params,
            metadata= self._job_metadata(circuit=circuit, meas_map=meas_map),
            **kwargs
        )

        logger.info(f"Submitted job with id '{job.id()}' for circuit '{circuit.name}':")
        logger.info(input_data)

        return job

    def retrieve_job(self, job_id) -> AzureQuantumJob:
        """ Returns the Job instance associated with the given id."""
        return self._provider.get_job(job_id)


class IonQSimulatorBackend(IonQBackend):
    backend_names = ("ionq.simulator",)

    def __init__(
        self,
        name: str,
        provider: "AzureQuantumProvider",
        **kwargs
    ):
        """Base class for interfacing with an IonQ Simulator backend"""
        default_config = BackendConfiguration.from_dict(
            {
                "backend_name": name,
                "backend_version": __version__,
                "simulator": True,
                "local": False,
                "coupling_map": None,
                "description": "IonQ simulator on Azure Quantum",
                "basis_gates": ionq_basis_gates,
                "memory": False,
                "n_qubits": 29,
                "conditional": False,
                "max_shots": 1,
                "max_experiments": 1,
                "open_pulse": False,
                "gates": [{"name": "TODO", "parameters": [], "qasm_def": "TODO"}],
            }
        )
        logger.info("Initializing IonQSimulatorBackend")
        configuration: BackendConfiguration = kwargs.pop("configuration", default_config)
        super().__init__(configuration=configuration, provider=provider, **kwargs)


class IonQQPUBackend(IonQBackend):
    backend_names = ("ionq.qpu",)

    def __init__(
        self,
        name: str,
        provider: "AzureQuantumProvider",
        **kwargs
    ):
        """Base class for interfacing with an IonQ QPU backend"""
        default_config = BackendConfiguration.from_dict(
            {
                "backend_name": name,
                "backend_version": __version__,
                "simulator": False,
                "local": False,
                "coupling_map": None,
                "description": "IonQ QPU on Azure Quantum",
                "basis_gates": ionq_basis_gates,
                "memory": False,
                "n_qubits": 11,
                "conditional": False,
                "max_shots": 10000,
                "max_experiments": 1,
                "open_pulse": False,
                "gates": [{"name": "TODO", "parameters": [], "qasm_def": "TODO"}],
            }
        )
        logger.info("Initializing IonQQPUBackend")
        configuration: BackendConfiguration = kwargs.pop("configuration", default_config)
        super().__init__(configuration=configuration, provider=provider, **kwargs)



import logging
from typing import TYPE_CHECKING
from azure.quantum.optimization import Problem

logger = logging.getLogger(__name__)

__all__ = ["OnlineProblem"]

if TYPE_CHECKING:
    from azure.quantum.workspace import Workspace


class OnlineProblem(object):
    def __init__(self, name: str, blob_uri: str, **kw):
        super(OnlineProblem, self).__init__(**kw)
        self.name = name
        self.uploaded_blob_uri = blob_uri

    def download(self, workspace: "Workspace") -> Problem:
        logger.warning("The problem will be downloaded to the client")
        return Problem.download(self, workspace)



# coding=utf-8
# --------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for license information.
# Code generated by Microsoft (R) AutoRest Code Generator.
# Changes may cause incorrect behavior and will be lost if the code is regenerated.
# --------------------------------------------------------------------------

from azure.core.exceptions import HttpResponseError
import msrest.serialization


class BlobDetails(msrest.serialization.Model):
    """Blob details.

    All required parameters must be populated in order to send to Azure.

    :ivar container_name: Required. The container name.
    :vartype container_name: str
    :ivar blob_name: The blob name.
    :vartype blob_name: str
    """

    _validation = {
        'container_name': {'required': True},
    }

    _attribute_map = {
        'container_name': {'key': 'containerName', 'type': 'str'},
        'blob_name': {'key': 'blobName', 'type': 'str'},
    }

    def __init__(
        self,
        **kwargs
    ):
        """
        :keyword container_name: Required. The container name.
        :paramtype container_name: str
        :keyword blob_name: The blob name.
        :paramtype blob_name: str
        """
        super(BlobDetails, self).__init__(**kwargs)
        self.container_name = kwargs['container_name']
        self.blob_name = kwargs.get('blob_name', None)


class CostEstimate(msrest.serialization.Model):
    """The job cost billed by the provider. The final cost on your bill might be slightly different due to added taxes and currency conversion rates.

    :ivar currency_code: The currency code.
    :vartype currency_code: str
    :ivar events: List of usage events.
    :vartype events: list[~azure.quantum._client.models.UsageEvent]
    :ivar estimated_total: The estimated total.
    :vartype estimated_total: float
    """

    _attribute_map = {
        'currency_code': {'key': 'currencyCode', 'type': 'str'},
        'events': {'key': 'events', 'type': '[UsageEvent]'},
        'estimated_total': {'key': 'estimatedTotal', 'type': 'float'},
    }

    def __init__(
        self,
        **kwargs
    ):
        """
        :keyword currency_code: The currency code.
        :paramtype currency_code: str
        :keyword events: List of usage events.
        :paramtype events: list[~azure.quantum._client.models.UsageEvent]
        :keyword estimated_total: The estimated total.
        :paramtype estimated_total: float
        """
        super(CostEstimate, self).__init__(**kwargs)
        self.currency_code = kwargs.get('currency_code', None)
        self.events = kwargs.get('events', None)
        self.estimated_total = kwargs.get('estimated_total', None)


class ErrorData(msrest.serialization.Model):
    """An error response from Azure.

    :ivar code: An identifier for the error. Codes are invariant and are intended to be consumed
     programmatically.
    :vartype code: str
    :ivar message: A message describing the error, intended to be suitable for displaying in a user
     interface.
    :vartype message: str
    """

    _attribute_map = {
        'code': {'key': 'code', 'type': 'str'},
        'message': {'key': 'message', 'type': 'str'},
    }

    def __init__(
        self,
        **kwargs
    ):
        """
        :keyword code: An identifier for the error. Codes are invariant and are intended to be consumed
         programmatically.
        :paramtype code: str
        :keyword message: A message describing the error, intended to be suitable for displaying in a
         user interface.
        :paramtype message: str
        """
        super(ErrorData, self).__init__(**kwargs)
        self.code = kwargs.get('code', None)
        self.message = kwargs.get('message', None)


class JobDetails(msrest.serialization.Model):
    """Job details.

    Variables are only populated by the server, and will be ignored when sending a request.

    All required parameters must be populated in order to send to Azure.

    :ivar id: The job id.
    :vartype id: str
    :ivar name: The job name. Is not required for the name to be unique and it's only used for
     display purposes.
    :vartype name: str
    :ivar container_uri: Required. The blob container SAS uri, the container is used to host job
     data.
    :vartype container_uri: str
    :ivar input_data_uri: The input blob SAS uri, if specified, it will override the default input
     blob in the container.
    :vartype input_data_uri: str
    :ivar input_data_format: Required. The format of the input data.
    :vartype input_data_format: str
    :ivar input_params: The input parameters for the job. JSON object used by the target solver. It
     is expected that the size of this object is small and only used to specify parameters for the
     execution target, not the input data.
    :vartype input_params: any
    :ivar provider_id: Required. The unique identifier for the provider.
    :vartype provider_id: str
    :ivar target: Required. The target identifier to run the job.
    :vartype target: str
    :ivar metadata: The job metadata. Metadata provides client the ability to store client-specific
     information.
    :vartype metadata: dict[str, str]
    :ivar output_data_uri: The output blob SAS uri. When a job finishes successfully, results will
     be uploaded to this blob.
    :vartype output_data_uri: str
    :ivar output_data_format: The format of the output data.
    :vartype output_data_format: str
    :ivar status: The job status. Possible values include: "Waiting", "Executing", "Succeeded",
     "Failed", "Cancelled".
    :vartype status: str or ~azure.quantum._client.models.JobStatus
    :ivar creation_time: The creation time of the job.
    :vartype creation_time: ~datetime.datetime
    :ivar begin_execution_time: The time when the job began execution.
    :vartype begin_execution_time: ~datetime.datetime
    :ivar end_execution_time: The time when the job finished execution.
    :vartype end_execution_time: ~datetime.datetime
    :ivar cancellation_time: The time when a job was successfully cancelled.
    :vartype cancellation_time: ~datetime.datetime
    :ivar cost_estimate: The job cost billed by the provider. The final cost on your bill might be
     slightly different due to added taxes and currency conversion rates.
    :vartype cost_estimate: ~azure.quantum._client.models.CostEstimate
    :ivar error_data: The error data for the job. This is expected only when Status 'Failed'.
    :vartype error_data: ~azure.quantum._client.models.ErrorData
    :ivar tags: A set of tags. List of user-supplied tags associated with the job.
    :vartype tags: list[str]
    """

    _validation = {
        'container_uri': {'required': True},
        'input_data_format': {'required': True},
        'provider_id': {'required': True},
        'target': {'required': True},
        'status': {'readonly': True},
        'creation_time': {'readonly': True},
        'begin_execution_time': {'readonly': True},
        'end_execution_time': {'readonly': True},
        'cancellation_time': {'readonly': True},
        'cost_estimate': {'readonly': True},
        'error_data': {'readonly': True},
    }

    _attribute_map = {
        'id': {'key': 'id', 'type': 'str'},
        'name': {'key': 'name', 'type': 'str'},
        'container_uri': {'key': 'containerUri', 'type': 'str'},
        'input_data_uri': {'key': 'inputDataUri', 'type': 'str'},
        'input_data_format': {'key': 'inputDataFormat', 'type': 'str'},
        'input_params': {'key': 'inputParams', 'type': 'object'},
        'provider_id': {'key': 'providerId', 'type': 'str'},
        'target': {'key': 'target', 'type': 'str'},
        'metadata': {'key': 'metadata', 'type': '{str}'},
        'output_data_uri': {'key': 'outputDataUri', 'type': 'str'},
        'output_data_format': {'key': 'outputDataFormat', 'type': 'str'},
        'status': {'key': 'status', 'type': 'str'},
        'creation_time': {'key': 'creationTime', 'type': 'iso-8601'},
        'begin_execution_time': {'key': 'beginExecutionTime', 'type': 'iso-8601'},
        'end_execution_time': {'key': 'endExecutionTime', 'type': 'iso-8601'},
        'cancellation_time': {'key': 'cancellationTime', 'type': 'iso-8601'},
        'cost_estimate': {'key': 'costEstimate', 'type': 'CostEstimate'},
        'error_data': {'key': 'errorData', 'type': 'ErrorData'},
        'tags': {'key': 'tags', 'type': '[str]'},
    }

    def __init__(
        self,
        **kwargs
    ):
        """
        :keyword id: The job id.
        :paramtype id: str
        :keyword name: The job name. Is not required for the name to be unique and it's only used for
         display purposes.
        :paramtype name: str
        :keyword container_uri: Required. The blob container SAS uri, the container is used to host job
         data.
        :paramtype container_uri: str
        :keyword input_data_uri: The input blob SAS uri, if specified, it will override the default
         input blob in the container.
        :paramtype input_data_uri: str
        :keyword input_data_format: Required. The format of the input data.
        :paramtype input_data_format: str
        :keyword input_params: The input parameters for the job. JSON object used by the target solver.
         It is expected that the size of this object is small and only used to specify parameters for
         the execution target, not the input data.
        :paramtype input_params: any
        :keyword provider_id: Required. The unique identifier for the provider.
        :paramtype provider_id: str
        :keyword target: Required. The target identifier to run the job.
        :paramtype target: str
        :keyword metadata: The job metadata. Metadata provides client the ability to store
         client-specific information.
        :paramtype metadata: dict[str, str]
        :keyword output_data_uri: The output blob SAS uri. When a job finishes successfully, results
         will be uploaded to this blob.
        :paramtype output_data_uri: str
        :keyword output_data_format: The format of the output data.
        :paramtype output_data_format: str
        :keyword tags: A set of tags. List of user-supplied tags associated with the job.
        :paramtype tags: list[str]
        """
        super(JobDetails, self).__init__(**kwargs)
        self.id = kwargs.get('id', None)
        self.name = kwargs.get('name', None)
        self.container_uri = kwargs['container_uri']
        self.input_data_uri = kwargs.get('input_data_uri', None)
        self.input_data_format = kwargs['input_data_format']
        self.input_params = kwargs.get('input_params', None)
        self.provider_id = kwargs['provider_id']
        self.target = kwargs['target']
        self.metadata = kwargs.get('metadata', None)
        self.output_data_uri = kwargs.get('output_data_uri', None)
        self.output_data_format = kwargs.get('output_data_format', None)
        self.status = None
        self.creation_time = None
        self.begin_execution_time = None
        self.end_execution_time = None
        self.cancellation_time = None
        self.cost_estimate = None
        self.error_data = None
        self.tags = kwargs.get('tags', None)


class JobDetailsList(msrest.serialization.Model):
    """List of job details.

    Variables are only populated by the server, and will be ignored when sending a request.

    :ivar value:
    :vartype value: list[~azure.quantum._client.models.JobDetails]
    :ivar count: Total records count number.
    :vartype count: long
    :ivar next_link: Link to the next page of results.
    :vartype next_link: str
    """

    _validation = {
        'value': {'readonly': True},
        'next_link': {'readonly': True},
    }

    _attribute_map = {
        'value': {'key': 'value', 'type': '[JobDetails]'},
        'count': {'key': 'count', 'type': 'long'},
        'next_link': {'key': 'nextLink', 'type': 'str'},
    }

    def __init__(
        self,
        **kwargs
    ):
        """
        :keyword count: Total records count number.
        :paramtype count: long
        """
        super(JobDetailsList, self).__init__(**kwargs)
        self.value = None
        self.count = kwargs.get('count', None)
        self.next_link = None


class JsonPatchDocument(msrest.serialization.Model):
    """A JSONPatch document as defined by RFC 6902.

    All required parameters must be populated in order to send to Azure.

    :ivar op: Required. The operation to be performed. Possible values include: "add", "remove",
     "replace", "move", "copy", "test".
    :vartype op: str or ~azure.quantum._client.models.JsonPatchOperation
    :ivar path: Required. A JSON-Pointer.
    :vartype path: str
    :ivar value: A value to be used in the operation on the path.
    :vartype value: any
    :ivar from_property: Optional field used in copy and move operations.
    :vartype from_property: str
    """

    _validation = {
        'op': {'required': True},
        'path': {'required': True},
    }

    _attribute_map = {
        'op': {'key': 'op', 'type': 'str'},
        'path': {'key': 'path', 'type': 'str'},
        'value': {'key': 'value', 'type': 'object'},
        'from_property': {'key': 'from', 'type': 'str'},
    }

    def __init__(
        self,
        **kwargs
    ):
        """
        :keyword op: Required. The operation to be performed. Possible values include: "add", "remove",
         "replace", "move", "copy", "test".
        :paramtype op: str or ~azure.quantum._client.models.JsonPatchOperation
        :keyword path: Required. A JSON-Pointer.
        :paramtype path: str
        :keyword value: A value to be used in the operation on the path.
        :paramtype value: any
        :keyword from_property: Optional field used in copy and move operations.
        :paramtype from_property: str
        """
        super(JsonPatchDocument, self).__init__(**kwargs)
        self.op = kwargs['op']
        self.path = kwargs['path']
        self.value = kwargs.get('value', None)
        self.from_property = kwargs.get('from_property', None)


class ProviderStatus(msrest.serialization.Model):
    """Providers status.

    Variables are only populated by the server, and will be ignored when sending a request.

    :ivar id: Provider id.
    :vartype id: str
    :ivar current_availability: Provider availability. Possible values include: "Available",
     "Degraded", "Unavailable".
    :vartype current_availability: str or ~azure.quantum._client.models.ProviderAvailability
    :ivar targets:
    :vartype targets: list[~azure.quantum._client.models.TargetStatus]
    """

    _validation = {
        'id': {'readonly': True},
        'current_availability': {'readonly': True},
        'targets': {'readonly': True},
    }

    _attribute_map = {
        'id': {'key': 'id', 'type': 'str'},
        'current_availability': {'key': 'currentAvailability', 'type': 'str'},
        'targets': {'key': 'targets', 'type': '[TargetStatus]'},
    }

    def __init__(
        self,
        **kwargs
    ):
        """
        """
        super(ProviderStatus, self).__init__(**kwargs)
        self.id = None
        self.current_availability = None
        self.targets = None


class ProviderStatusList(msrest.serialization.Model):
    """Providers status.

    Variables are only populated by the server, and will be ignored when sending a request.

    :ivar value:
    :vartype value: list[~azure.quantum._client.models.ProviderStatus]
    :ivar next_link: Link to the next page of results.
    :vartype next_link: str
    """

    _validation = {
        'value': {'readonly': True},
        'next_link': {'readonly': True},
    }

    _attribute_map = {
        'value': {'key': 'value', 'type': '[ProviderStatus]'},
        'next_link': {'key': 'nextLink', 'type': 'str'},
    }

    def __init__(
        self,
        **kwargs
    ):
        """
        """
        super(ProviderStatusList, self).__init__(**kwargs)
        self.value = None
        self.next_link = None


class Quota(msrest.serialization.Model):
    """Quota information.

    :ivar dimension: The name of the dimension associated with the quota.
    :vartype dimension: str
    :ivar scope: The scope at which the quota is applied. Possible values include: "Workspace",
     "Subscription".
    :vartype scope: str or ~azure.quantum._client.models.DimensionScope
    :ivar provider_id: The unique identifier for the provider.
    :vartype provider_id: str
    :ivar utilization: The amount of the usage that has been applied for the current period.
    :vartype utilization: float
    :ivar holds: The amount of the usage that has been reserved but not applied for the current
     period.
    :vartype holds: float
    :ivar limit: The maximum amount of usage allowed for the current period.
    :vartype limit: float
    :ivar period: The time period in which the quota's underlying meter is accumulated. Based on
     calendar year. 'None' is used for concurrent quotas. Possible values include: "None",
     "Monthly".
    :vartype period: str or ~azure.quantum._client.models.MeterPeriod
    """

    _attribute_map = {
        'dimension': {'key': 'dimension', 'type': 'str'},
        'scope': {'key': 'scope', 'type': 'str'},
        'provider_id': {'key': 'providerId', 'type': 'str'},
        'utilization': {'key': 'utilization', 'type': 'float'},
        'holds': {'key': 'holds', 'type': 'float'},
        'limit': {'key': 'limit', 'type': 'float'},
        'period': {'key': 'period', 'type': 'str'},
    }

    def __init__(
        self,
        **kwargs
    ):
        """
        :keyword dimension: The name of the dimension associated with the quota.
        :paramtype dimension: str
        :keyword scope: The scope at which the quota is applied. Possible values include: "Workspace",
         "Subscription".
        :paramtype scope: str or ~azure.quantum._client.models.DimensionScope
        :keyword provider_id: The unique identifier for the provider.
        :paramtype provider_id: str
        :keyword utilization: The amount of the usage that has been applied for the current period.
        :paramtype utilization: float
        :keyword holds: The amount of the usage that has been reserved but not applied for the current
         period.
        :paramtype holds: float
        :keyword limit: The maximum amount of usage allowed for the current period.
        :paramtype limit: float
        :keyword period: The time period in which the quota's underlying meter is accumulated. Based on
         calendar year. 'None' is used for concurrent quotas. Possible values include: "None",
         "Monthly".
        :paramtype period: str or ~azure.quantum._client.models.MeterPeriod
        """
        super(Quota, self).__init__(**kwargs)
        self.dimension = kwargs.get('dimension', None)
        self.scope = kwargs.get('scope', None)
        self.provider_id = kwargs.get('provider_id', None)
        self.utilization = kwargs.get('utilization', None)
        self.holds = kwargs.get('holds', None)
        self.limit = kwargs.get('limit', None)
        self.period = kwargs.get('period', None)


class QuotaList(msrest.serialization.Model):
    """List of quotas.

    Variables are only populated by the server, and will be ignored when sending a request.

    :ivar value:
    :vartype value: list[~azure.quantum._client.models.Quota]
    :ivar next_link: Link to the next page of results.
    :vartype next_link: str
    """

    _validation = {
        'value': {'readonly': True},
        'next_link': {'readonly': True},
    }

    _attribute_map = {
        'value': {'key': 'value', 'type': '[Quota]'},
        'next_link': {'key': 'nextLink', 'type': 'str'},
    }

    def __init__(
        self,
        **kwargs
    ):
        """
        """
        super(QuotaList, self).__init__(**kwargs)
        self.value = None
        self.next_link = None


class RestError(msrest.serialization.Model):
    """Error information returned by the API.

    :ivar error: An error response from Azure.
    :vartype error: ~azure.quantum._client.models.ErrorData
    """

    _attribute_map = {
        'error': {'key': 'error', 'type': 'ErrorData'},
    }

    def __init__(
        self,
        **kwargs
    ):
        """
        :keyword error: An error response from Azure.
        :paramtype error: ~azure.quantum._client.models.ErrorData
        """
        super(RestError, self).__init__(**kwargs)
        self.error = kwargs.get('error', None)


class SasUriResponse(msrest.serialization.Model):
    """Get SAS URL operation response.

    :ivar sas_uri: A URL with a SAS token to upload a blob for execution in the given workspace.
    :vartype sas_uri: str
    """

    _attribute_map = {
        'sas_uri': {'key': 'sasUri', 'type': 'str'},
    }

    def __init__(
        self,
        **kwargs
    ):
        """
        :keyword sas_uri: A URL with a SAS token to upload a blob for execution in the given workspace.
        :paramtype sas_uri: str
        """
        super(SasUriResponse, self).__init__(**kwargs)
        self.sas_uri = kwargs.get('sas_uri', None)


class TargetStatus(msrest.serialization.Model):
    """Target status.

    Variables are only populated by the server, and will be ignored when sending a request.

    :ivar id: Target id.
    :vartype id: str
    :ivar current_availability: Target availability. Possible values include: "Available",
     "Degraded", "Unavailable".
    :vartype current_availability: str or ~azure.quantum._client.models.TargetAvailability
    :ivar average_queue_time: Average queue time in seconds.
    :vartype average_queue_time: long
    :ivar status_page: A page with detailed status of the provider.
    :vartype status_page: str
    """

    _validation = {
        'id': {'readonly': True},
        'current_availability': {'readonly': True},
        'average_queue_time': {'readonly': True},
        'status_page': {'readonly': True},
    }

    _attribute_map = {
        'id': {'key': 'id', 'type': 'str'},
        'current_availability': {'key': 'currentAvailability', 'type': 'str'},
        'average_queue_time': {'key': 'averageQueueTime', 'type': 'long'},
        'status_page': {'key': 'statusPage', 'type': 'str'},
    }

    def __init__(
        self,
        **kwargs
    ):
        """
        """
        super(TargetStatus, self).__init__(**kwargs)
        self.id = None
        self.current_availability = None
        self.average_queue_time = None
        self.status_page = None


class UsageEvent(msrest.serialization.Model):
    """Usage event details.

    :ivar dimension_id: The dimension id.
    :vartype dimension_id: str
    :ivar dimension_name: The dimension name.
    :vartype dimension_name: str
    :ivar measure_unit: The unit of measure.
    :vartype measure_unit: str
    :ivar amount_billed: The amount billed.
    :vartype amount_billed: float
    :ivar amount_consumed: The amount consumed.
    :vartype amount_consumed: float
    :ivar unit_price: The unit price.
    :vartype unit_price: float
    """

    _attribute_map = {
        'dimension_id': {'key': 'dimensionId', 'type': 'str'},
        'dimension_name': {'key': 'dimensionName', 'type': 'str'},
        'measure_unit': {'key': 'measureUnit', 'type': 'str'},
        'amount_billed': {'key': 'amountBilled', 'type': 'float'},
        'amount_consumed': {'key': 'amountConsumed', 'type': 'float'},
        'unit_price': {'key': 'unitPrice', 'type': 'float'},
    }

    def __init__(
        self,
        **kwargs
    ):
        """
        :keyword dimension_id: The dimension id.
        :paramtype dimension_id: str
        :keyword dimension_name: The dimension name.
        :paramtype dimension_name: str
        :keyword measure_unit: The unit of measure.
        :paramtype measure_unit: str
        :keyword amount_billed: The amount billed.
        :paramtype amount_billed: float
        :keyword amount_consumed: The amount consumed.
        :paramtype amount_consumed: float
        :keyword unit_price: The unit price.
        :paramtype unit_price: float
        """
        super(UsageEvent, self).__init__(**kwargs)
        self.dimension_id = kwargs.get('dimension_id', None)
        self.dimension_name = kwargs.get('dimension_name', None)
        self.measure_unit = kwargs.get('measure_unit', None)
        self.amount_billed = kwargs.get('amount_billed', None)
        self.amount_consumed = kwargs.get('amount_consumed', None)
        self.unit_price = kwargs.get('unit_price', None)



##
# Copyright (c) Microsoft Corporation.
# Licensed under the MIT License.
##
import numpy as np

from typing import TYPE_CHECKING, Any, Dict, Sequence

from azure.quantum.cirq.targets.target import Target as CirqTarget
from azure.quantum.cirq.job import Job as CirqJob
from azure.quantum.target.quantinuum import Quantinuum

if TYPE_CHECKING:
    import cirq
    from azure.quantum import Workspace
    from azure.quantum import Job as AzureJob


class QuantinuumTarget(Quantinuum, CirqTarget):
    """Base class for interfacing with an Quantinuum (formerly Honeywell) backend in Azure Quantum"""

    def __init__(
        self,
        workspace: "Workspace",
        name: str,
        input_data_format: str = "honeywell.openqasm.v1",
        output_data_format: str = "honeywell.quantum-results.v1",
        provider_id: str = "quantinuum",
        content_type: str = "application/qasm",
        encoding: str = "",
        **kwargs
    ):
        super().__init__(
            workspace=workspace,
            name=name,
            input_data_format=input_data_format,
            output_data_format=output_data_format,
            provider_id=provider_id,
            content_type=content_type,
            encoding=encoding,
            **kwargs
        )

    @staticmethod
    def _translate_cirq_circuit(circuit) -> str:
        """Translate `cirq` circuit to OpenQASM 2.0."""
        return circuit.to_qasm()

    @staticmethod
    def _to_cirq_result(result: Dict[str, Any], param_resolver, **kwargs):
        from cirq import Result
        measurements = {
            key.lstrip("m_"): np.array([[int(_v)] for _v in value])
            for key, value in result.items()
            if key.startswith("m_")
        }
        return Result(params=param_resolver, measurements=measurements)

    def _to_cirq_job(self, azure_job: "AzureJob", program: "cirq.Circuit" = None):
        """Convert Azure job to Cirq job"""
        if "measurement_dict" not in azure_job.details.metadata and program is None:
            raise ValueError("Parameter 'measurement_dict' not found in job metadata.")
        measurement_dict = azure_job.details.metadata.get("measurement_dict")
        return CirqJob(azure_job=azure_job, program=program, measurement_dict=measurement_dict)

    @staticmethod
    def _measurement_dict(program) -> Dict[str, Sequence[int]]:
        """Returns a dictionary of measurement keys to target qubit index."""
        from cirq import MeasurementGate
        measurements = [
            op for op in program.all_operations() if isinstance(op.gate, MeasurementGate)
        ]
        return {
            meas.gate.key: [q.x for q in meas.qubits] for meas in measurements
        }

    def estimate_cost(
        self,
        program: str,
        repetitions: int
    ) -> float:
        """Estimate cost for running this program

        :param program: Cirq quantum program
        :type program: str, optional
        :param repetitions: Number of repetitions
        :type repetitions: int, optional
        :return: Price estimate in HQC
        :rtype: float
        """
        serialized_program = self._translate_circuit(program)
        return super().estimate_cost(
            circuit=serialized_program,
            num_shots=repetitions
        )

    def submit(
        self,
        program: "cirq.Circuit",
        name: str = "cirq-job",
        repetitions: int = 500,
        **kwargs
    ) -> "CirqJob":
        """Submit a Cirq quantum circuit

        :param program: Quantum program
        :type program: cirq.Circuit
        :param name: Job name
        :type name: str
        :param repetitions: Number of shots, defaults to
            provider default value
        :type repetitions: int
        :return: Azure Quantum job
        :rtype: Job
        """
        serialized_program = self._translate_circuit(program)
        metadata = {
            "qubits": len(program.all_qubits()),
            "repetitions": repetitions,
            "measurement_dict": self._measurement_dict(program)
        }
        # Override metadata with value from kwargs
        metadata.update(kwargs.get("metadata", {}))
        azure_job = super().submit(
            circuit=serialized_program,
            name=name,
            num_shots=repetitions,
            metadata=metadata,
            **kwargs
        )
        return self._to_cirq_job(azure_job=azure_job)



##
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License.
##
import logging
from typing import TYPE_CHECKING, Union, Optional
from azure.quantum import Workspace
from azure.quantum.target import Solver
from azure.quantum import Job

if TYPE_CHECKING:
    from azure.quantum.optimization import Problem

logger = logging.getLogger(__name__)
__all__ = ["SimulatedBifurcationMachine"]


class SimulatedBifurcationMachine(Solver):
    target_names = (
        "toshiba.sbm.ising",
    )
    def __init__(
        self,
        workspace: Workspace,
        name="toshiba.sbm.ising",
        steps: Optional[int] = None,
        loops: Optional[int] = None,
        target: Optional[float] = None,
        maxout: Optional[int] = None,
        timeout: Optional[float] = None,
        dt: Optional[float] = None,
        C: Optional[float] = None,
        algo: Optional[str] = None,
        auto: Optional[bool] = None,
        **kwargs
    ):
        """The constructor of Toshiba's simulated bifurcation machine.

        This solver enables users to quickly obtain nearly optimal
        solutions for large combinatorial optimization problems.
        It is developed based on the theory described in the following paper:
        Goto, H., Tatsumura, K., & Dixon, A. R. (2019)
        Combinatorial optimization by simulating adiabatic bifurcations
        in nonlinear Hamiltonian systems, Science Advances,
        5(4), DOI:10.1126/sciadv.aav2372
        :param steps:
           The number of steps in a computation request.
           The default is 0.
           The value 0 means auto step where SBM computation
           service dynamically determines the number of steps.
           The maximum is 100,000,000.
        :param loops:
            The number of loops in SBM computation.
            SBM computation service searches for a better solution
            while repeating loops as many times as is specified.
            The default is 1. If 0 (zero) is specified, computation
            will be repeated until a timeout occurs. The maximum is 10,000,000.
        :param target:
            The end condition of a computation request. When the
            evaluation value reaches this value, the computation will stop.
            If 0 is specified for the parameter loops,
            loops will be repeated either until
            the objective function reaches the value
            specified in the parameter target or until a timeout occurs.
        :param maxout:
            The upper limit of the number of solutions
            to be outputted. The default is 1.
            Until the limit specified by maxout is reached,
            SBM computation service outputs the
            obtained solutions in descending order
            of the value of the objective function.
            The maximum is 1,000.
        :param timeout:
            The maximum computation time (timeout) in seconds.
            The default is 10.
            When the computation time reaches the upper limit before
            completing the computation for steps * loops,
            the computation ends at that point.
            In this case, the execution result will be the
            best solution obtained thus far.
            If 0 is specified for the parameter loops, loops will be repeated
            until the time specified by the parameter timeout expires.
            The maximum is 360.
        :param dt:
            The time per step. The range of the value is greater than
            0.0 and less than or equal to 1.0.
        :param C:
            Corresponds to the constant ξ0,
            appearing in the paper by Goto, Tatsumura,
            & Dixon (2019, p. 2), which is the theoretical basis of SBM.
            Specify the constant as a single-precision floating point
            number, equal to or more than 0.
            If the value is 0, the value C is automatically determined.
        :param algo:
            The type of SBM computation algorithm.
            The default is "2.0". One of "1.5" or "2.0".
            Depending on the type of problem, there may be differences
            in performance between the "1.5" and "2.0" algorithms.
            Try both and decide which yields better performance.
        :param auto:
            The parameter auto tuning flag. The default is "false."
            If the value is "true," SBM computation service searches for
            the values of the parameters automatically to obtain
            the best solution.
            Parameters other than `auto` are treated as follows
            in this case. `algo` and `dt` are ignored and tuned automatically.
            `loops` and `maxout` are ignored.
            `timeout` can be specified as the total computation time (sec).
            Other parameters are treated as defined
        """
        super().__init__(
            workspace=workspace,
            provider_id="toshiba",
            name=name,
            input_data_format="microsoft.qio.v2",
            output_data_format="microsoft.qio-results.v2",
            nested_params=False,
            force_str_params=False,
            **kwargs
        )
        self.set_one_param("steps", steps)
        self.set_one_param("loops", loops)
        self.set_one_param("target", target)
        self.set_one_param("maxout", maxout)
        self.set_one_param("timeout", timeout)
        self.set_one_param("dt", dt)
        self.set_one_param("C", C)
        self.set_one_param("algo", algo)
        self.set_one_param("auto", auto)




##
# Copyright (c) Microsoft Corporation.
# Licensed under the MIT License.
##

from azure.quantum.qiskit.backends.ionq import (
    IonQBackend,
    IonQQPUBackend,
    IonQSimulatorBackend
)

from azure.quantum.qiskit.backends.honeywell import (
    HoneywellBackend,
    HoneywellQPUBackend,
    HoneywellAPIValidatorBackend,
    HoneywellSimulatorBackend
)

from azure.quantum.qiskit.backends.quantinuum import (
    QuantinuumBackend,
    QuantinuumQPUBackend,
    QuantinuumAPIValidatorBackend,
    QuantinuumSimulatorBackend
)

# Default targets to use when there is no target class
# associated with a given target ID
DEFAULT_TARGETS = {}



##
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License.
##
import logging

from typing import TYPE_CHECKING, Union, Any, Optional
from enum import Enum
from azure.quantum.job.base_job import ContentType
from azure.quantum import Workspace, Job
from azure.quantum.job.base_job import DEFAULT_TIMEOUT
from azure.quantum.target.target import Target

if TYPE_CHECKING:
    from azure.quantum.optimization.problem import Problem

logger = logging.getLogger(__name__)

__all__ = [
    "HardwarePlatform",
    "RangeSchedule",
    "Solver",
]


class HardwarePlatform(Enum):
    CPU = 1
    FPGA = 2


class RangeSchedule:
    def __init__(self, schedule_type: str, initial: float, final: float):
        """Constructor of RangeSchedule for solvers.

        :param schedule_type:
            str, type of the RangeSchedule
            currently only supports 'linear' and 'geometric'.
        :param initial:
            float, initial value of RangeSchedule
        :param final:
            float, final value of the RangeSchedule
        """
        self.schedule_type = schedule_type
        self.initial = initial
        self.final = final
        if not (
            self.schedule_type == "linear" or self.schedule_type == "geometric"
        ):
            raise ValueError(
                '"schedule_type" must be "linear" or "geometric"!'
            )

proto_valid_solver_names = [
    "PopulationAnnealing",
    "SubstochasticMonteCarlo"
]

class Solver(Target):
    def __init__(
        self,
        workspace: Workspace,
        name: str,
        provider_id: str = "Microsoft",
        input_data_format="microsoft.qio.v2",
        output_data_format="microsoft.qio-results.v2",
        nested_params: bool = True,
        force_str_params: bool = False,
        params: dict = None,
        content_type : Optional[ContentType] = ContentType.json,
        **kwargs
    ):
        self.provider_id = provider_id
        self.nested_params = nested_params
        self.force_str_params = force_str_params
        self.params = {"params": {}} if nested_params else {}
        params = params or {}
        name = self._switch_name(name, **params)

        super().__init__(
            workspace=workspace,
            name=name,
            input_data_format=input_data_format,
            output_data_format=output_data_format,
            provider_id=provider_id,
            content_type = content_type,
            encoding="gzip",
            **kwargs
        )
 
        self._set_params(**params)
        self._check_params(**params)


    """Constants that define thresholds for submission warnings
    """
    SWEEPS_WARNING = 10000
    TIMEOUT_WARNING = 600

    def _switch_name(self, name, **params):
        return name

    def _check_params(self, **params):
        pass

    @staticmethod
    def _encode_input_data(data: "Problem") -> bytes:
        return data.to_blob()

    def submit(
        self, problem: Union[str, "Problem"]) -> Job:
        """Submits a job to execution to the associated
        Azure Quantum Workspace.

        :param problem:
            The Problem to solve. It can be an instance of a Problem,
            or the URL of an Azure Storage Blob where the serialized version
            of a Problem has been uploaded.
        """
        
        from azure.quantum.optimization import Problem
        if isinstance(problem, Problem):
            self.check_valid_problem(problem)
            return super().submit(
                input_data=problem,
                name=problem.name,
                input_params=self.params,
                blob_name="inputData",
                content_type = problem.content_type
            )

        else:
            if hasattr(problem, "uploaded_blob_uri"):
                name = problem.name
                problem_uri = problem.uploaded_blob_uri

            elif isinstance(problem, str):
                name = "Optimization problem"
                problem_uri = problem
            
            else:
                raise ValueError("Cannot submit problem: should be of type str, Problem or have uploaded_blob_uri attribute.")

            # Create job from storage URI
            job = Job.from_storage_uri(
                workspace=self.workspace,
                name=name,
                target=self.name,
                input_data_uri=problem_uri,
                provider_id=self.provider_id,
                input_data_format=self.input_data_format,
                output_data_format=self.output_data_format,
                input_params=self.params
            )

        return job

    def optimize(self, problem: Union[str, "Problem"], timeout_secs: int=DEFAULT_TIMEOUT):
        """[Submits the Problem to the associated
            Azure Quantum Workspace and get the results.

        :param problem:
            The Problem to solve. It can be an instance of a Problem,
            or the URL of an Azure Storage Blob where the serialized version
            of a Problem has been uploaded.
        :type problem: Union[str, Problem]
        :param timeout_secs: Timeout in seconds, defaults to 300
        :type timeout_secs: int
        :return: Job results
        :rtype: dict
        """
        if not isinstance(problem, str):
            self.check_submission_warnings(problem)

        job = self.submit(problem)
        logger.info(f"Submitted job: '{job.id}'")

        return job.get_results(timeout_secs=timeout_secs)

    def set_one_param(self, name: str, value: Any):
        if value is not None:
            params = (
                self.params["params"] if self.nested_params else self.params
            )
            params[name] = str(value) if self.force_str_params else value

    def _set_params(self, **params):
        for param_name, param in params.items():
            self.set_one_param(param_name, param)

    def set_number_of_solutions(self, number_of_solutions: int):
        """Sets the number of solutions to return.

        Applies to all solvers.
        Default value is 1 if not supplied.

        :param number_of_solutions:
            Number of solutions to return. Must be a positive integer.
        """
        self.set_one_param("number_of_solutions", number_of_solutions)

    def check_submission_warnings(self, problem: "Problem"):
        # print a warning if we suspect the job
        # may take long based on its configured properties.
        is_large_problem = problem.is_large()
        if is_large_problem:
            if self.nested_params and "sweeps" in self.params["params"]:
                sweeps = int(self.params["params"]["sweeps"])
                # if problem is large and sweeps is large, warn.
                if sweeps >= Solver.SWEEPS_WARNING:
                    logger.warning(
                        f"There is a large problem submitted and \
                        a large number of sweeps ({sweeps}) configured. \
                        This submission could result in a long runtime."
                    )

        # do a timeout check if param-free, to warn
        # new users who accidentally set high timeout values.
        if self.nested_params and "timeout" in self.params["params"]:
            timeout = int(self.params["params"]["timeout"])
            if timeout >= Solver.TIMEOUT_WARNING:
                logger.warning(
                    f"A large timeout has been set for this submission \
                        ({timeout}). \
                        If this is intended, disregard this warning. \
                        Otherwise, consider cancelling the job \
                        and resubmitting with a lower timeout."
                )

    def check_valid_problem(self, problem):
        # Evaluate Problem and Solver compatibility.
        from azure.quantum.optimization.problem import ProblemType
        if problem.problem_type in {ProblemType.pubo_grouped, ProblemType.ising_grouped}:
            if not self.supports_grouped_terms():
                raise ValueError(
                    f"Solver type is not compatible"
                    f"with problem type {problem.problem_type};"
                    f"Try PopulationAnnealing or SubstochasticMonteCarlo."
                )
        if problem.content_type == ContentType.protobuf:
            if not self.supports_protobuf() and self.name not in proto_valid_solver_names:
                raise ValueError(
                    f"Solver `{self.name} type is not compatible "
                    f"for serialization with protobuf; "
                    f"Try PopulationAnnealing or SubstochasticMonteCarlo."
                )

    def supports_grouped_terms(self):
        """
        Return whether or not the Solver class supported grouped terms in the cost function.
        This should be overridden by Solver subclasses which do support grouped terms.
        """
        return False
    
    class ScheduleEvolution(Enum):
        INCREASING = 1
        DECREASING = 2

    def check_set_schedule(
        self,
        schedule_name: str,
        schedule_value: RangeSchedule,
        evolution: Optional[ScheduleEvolution] = None,
        lower_bound_exclusive: Optional[float] = None,
        lower_bound_inclusive: Optional[float] = None,
    ):
        """Set the parameter `schedule_name` from the value `schedule_value`
        and check that it is of type `RangeSchedule` and each end satisfies
        the specified bound.

        :param schedule_name:
            Name of the schedule parameter.
        :param schedule_value:
            Schedule value to be assigned and checked.
        :param evolution
            Expected schedule evolution (INCREASING or DECREASING)
        :lower_bound_exclusive:
            Exclusive lower bound for both ends of the schedule, optional.
        :lower_bound_inclusive:
            Inclusive lower bound for both ends of the schedule, optional.
        """
        if not (schedule_value is None):
            if not isinstance(schedule_value, RangeSchedule):
                raise ValueError(
                    f"{schedule_name} must be of type RangeSchedule; found"
                    f" type({schedule_name})={type(schedule_value).__name__}."
                )
            schedule_param = {
                "type": schedule_value.schedule_type,
                "initial": schedule_value.initial,
                "final": schedule_value.final,
            }
            if evolution is not None:
                if (evolution == self.ScheduleEvolution.INCREASING and
                        schedule_value.initial > schedule_value.final):
                    raise ValueError(
                            f"Schedule for {schedule_name} must be increasing;"
                            f" found {schedule_name}.initial"
                            f"={schedule_value.initial}"
                            f" > {schedule_value.final}"
                            f"={schedule_name}.final."
                    )
                if (evolution == self.ScheduleEvolution.DECREASING and
                        schedule_value.initial < schedule_value.final):
                    raise ValueError(
                            f"Schedule for {schedule_name} must be decreasing;"
                            f" found {schedule_name}.initial"
                            f"={schedule_value.initial}"
                            f" < {schedule_value.final}"
                            f"={schedule_name}.final."
                    )
            self.check_limit(
                    f"{schedule_name}.initial",
                    schedule_value.initial,
                    lower_bound_exclusive=lower_bound_exclusive,
                    lower_bound_inclusive=lower_bound_inclusive)
            self.check_limit(
                    f"{schedule_name}.final",
                    schedule_value.final,
                    lower_bound_exclusive=lower_bound_exclusive,
                    lower_bound_inclusive=lower_bound_inclusive)
            self.set_one_param(schedule_name, schedule_param)

    def check_set_positive_int(
            self,
            parameter_name: str,
            parameter_value: int):
        """Set the parameter `parameter_name` from the value `parameter_value`
        and check that it is a positive integer.

        :param parameter_name:
            Name of the parameter.
        :param parameter_value:
            Value to be assigned and checked.
        """
        if not (parameter_value is None):
            if not isinstance(parameter_value, int):
                raise ValueError(
                        f"{parameter_name} must be of type int; found"
                        f"type({parameter_name})"
                        f"={type(parameter_value).__name__}.")
            if parameter_value <= 0:
                raise ValueError(
                        f"{parameter_name} must be positive; found "
                        f"{parameter_name}={parameter_value}.")
            self.set_one_param(parameter_name, parameter_value)

    def check_set_float(
        self,
        parameter_name: str,
        parameter_value: float,
        lower_bound_exclusive: Optional[float] = None,
        lower_bound_inclusive: Optional[float] = None,
    ):
        """Set the parameter `parameter_name` from the value `parameter_value`
        and check that it has a float value satisfying bounds.

        :param parameter_name:
            Name of the parameter.
        :param parameter_value:
            Value to be assigned and checked.
        :lower_bound_exclusive:
            Exclusive lower bound to check parameter_value against, optional.
        :lower_bound_inclusive:
            Inclusive lower bound to check parameter_value against, optional.
        """
        if not (parameter_value is None):
            if not (isinstance(parameter_value, float) or
                    isinstance(parameter_value, int)):
                raise ValueError(f"{parameter_name} must be a float!")
            else:
                self.check_limit(
                        parameter_name=parameter_name,
                        parameter_value=parameter_value,
                        lower_bound_exclusive=lower_bound_exclusive,
                        lower_bound_inclusive=lower_bound_inclusive)
                self.set_one_param(parameter_name, parameter_value)

    def check_limit(
        self,
        parameter_name: str,
        parameter_value: Optional[float],
        lower_bound_exclusive: Optional[float] = None,
        lower_bound_inclusive: Optional[float] = None,
    ):
        """Check whether `parameter_value` satisfies a lower bound.

        :param parameter_name:
            Name of the parameter.
        :param parameter_value:
            Value to be checked.
        :lower_bound_exclusive:
            Exclusive lower bound to check parameter_value against, optional.
        :lower_bound_inclusive:
            Inclusive lower bound to check parameter_value against, optional.
        """
        if not (parameter_value is None):
            if (lower_bound_exclusive is not None and
                    parameter_value <= lower_bound_exclusive):
                raise ValueError(
                    f"{parameter_name} must be greater than "
                    f"{lower_bound_exclusive}; "
                    f"found {parameter_name}={parameter_value}."
                )
            if (lower_bound_inclusive is not None and
                    parameter_value < lower_bound_inclusive):
                raise ValueError(
                    f"{parameter_name} must be greater equal "
                    f"{lower_bound_inclusive}; found "
                    f"{parameter_name}={parameter_value}."
                )
